# $Id: tml-mysql.rb,v 1.6 2004/04/24 14:06:51 tommy Exp $
#
# Copyright (C) 2003-2004 TOMITA Masahiro
# tommy@tmtm.org
#

require "mysql"

class TML; end

class TML::Mysql
  @@my = nil

  def self.my()
    unless @@my then
      @@my = Mysql.new(TML.param(:mysql_server), TML.param(:mysql_user), TML.param(:mysql_passwd), TML.param(:mysql_db))
    end
    @@my
  end

  class Lock
    @@locks = []

    def my() TML::Mysql.my end

    def self.callback(ident)
      proc do
        TML::Mysql.my.query("delete from locks where ident='#{TML::Mysql.my.quote ident}'")
        TML.log "unlock: #{ident}"
        @@locks.delete ident
      end
    end

    def initialize(ident)
      if @@locks.include? ident then
        raise "dead lock detected: #{ident}"
      end
      st = Time.now
      begin
        ts, = my.query("select unix_timestamp(timestamp) from locks where ident='#{my.quote ident}' limit 1").fetch_row
        if ts and ts.to_i < Time.now.to_i-TML.param(:max_lock_time) then
          my.query("delete from locks where ident='#{my.quote ident}'")
          TML.log "#{ident}: max_lock_time(#{TML.param(:max_lock_time)}) exceeded"
        end
        my.query("insert into locks (ident) values ('#{my.quote ident}')")
        ObjectSpace.define_finalizer(self, Lock.callback(ident))
        @ident = ident
        @@locks << ident
        TML.log "lock: #{ident}"
      rescue MysqlError => err
        if err.errno == MysqlError::ER_DUP_ENTRY then
          if st+TML.param(:lock_timeout) < Time.now then
            raise Error.new(:ETIMEDOUT, "lock timed out")
          end
          sleep 1
          retry
        else
          raise
        end
      end
    end

    def unlock()
      my.query("delete from locks where ident='#{my.quote @ident}'")
      TML.log "unlock: #{@ident}"
      ObjectSpace.undefine_finalizer(self)
      @@locks.delete @ident
    end
  end

  class Meta
    private

    def my() TML::Mysql.my end
    def q(s) s == nil ? "NULL" : "'#{my.quote(s.to_s)}'" end
    def alt_key() nil end
    def log_str() nil end

    def check_attr(attr)
      attr.keys.each do |k|
        unless items.include? k then
          TML::Error.new(:EINVAL, "#{attr}: no such attribute")
        end
      end
    end

    def where(k)
      unless alt_key then
        "where #{key}=#{q k}"
      else
        "where #{key}=#{q k} and #{alt_key}=#{q @alt_key}"
      end
    end

    public

    def have?(k)
      my.query("select 1 from #{table} #{where k}").fetch_row
    end

    def list()
      ret = []
      q = "select #{key} from #{table}"
      q += " where #{alt_key}=#{q @alt_key}" if alt_key
      my.query(q).each do |k,|
        ret << k
      end
      ret
    end

    def list_with_data()
      ret = []
      q = "select #{key},#{items.join(",")} from #{table}"
      q += " where #{alt_key}=#{q @alt_key}" if alt_key
      my.query(q).each_hash do |h|
        ret << h
      end
      ret
    end

    def add(k, attr={})
      begin
        check_attr attr

        keys = [key] + attr.keys
        keys << alt_key if alt_key
        keys_str = keys.join(",")

        values = [k] + attr.values
        values << @alt_key if alt_key
        values_str = values.map{|v| q(v)}.join(",")

        my.query("insert into #{table} (#{keys_str}) values (#{values_str})")
      rescue MysqlError => err
        if err.errno == MysqlError::ER_DUP_ENTRY then
          raise TML::Error.new(:EEXIST, "#{k}: already exist")
        end
        raise
      end
      TML.log "add #{log_str}: #{k}" if log_str
    end

    def delete(k)
      my.query("delete from #{table} #{where k}")
      if my.affected_rows == 0 then
        raise TML::Error.new(:ENOENT, "#{k}: not found")
      end
      TML.log "delete #{log_str}: #{k}" if log_str
    end

    def set(k, attr)
      unless have? k then
        raise TML::Error.new(:ENOENT, "#{k}: not found")
      end
      check_attr attr
      set = attr.map{|i,v|"#{i}=#{q v}"}.join(",")
      my.query("update #{table} set #{set} #{where k}")
      TML.log "set #{log_str}: #{k}" if log_str
    end

    def get(k)
      unless have? k then
        raise TML::Error.new(:ENOENT, "#{k}: not found")
      end
      if items.empty? then
        return {}
      end
      my.query("select #{items.join(",")} from #{table} #{where k} limit 1").fetch_hash
    end
  end

  class ML < Meta
    def table() "ml" end
    def key() "name" end
    def items()
      %w(sender overview not_member subject_mlname
	 counter_width remote_command remote_admin spool web_interface
	 need_confirm_command allowed_command anonymous_command admin_command
	 logging replyto_type
      )
    end
    def log_str() "ml" end
  end

  class Counter < Meta
    def table() "counter" end
    def key() "ml" end
    def items() ["counter"] end
    def log_str() "counter" end

    def add(ml)
      super(ml, {"counter"=>"0"})
    end

    def set(ml, cnt)
      super(ml, {"counter"=>cnt.to_s})
    end

    def get(ml)
      super(ml)["counter"].to_i
    end
  end

  class Passwd < Meta
    def table() "passwd" end
    def key() "addr" end
    def items() ["passwd"] end
    def log_str() "password" end

    def add(addr, pw)
      super(addr, {"passwd"=>pw})
    end

    def set(addr, pw)
      super(addr, {"passwd"=>pw})
    end

    def auth(addr, pw)
      begin
        p = get(addr)["passwd"]
        if p == pw then
          return true
        end
      rescue TML::Error => err
        raise unless err.id == :ENOENT
      end
      TML.log "auth failed: #{addr}"
      return false
    end
  end

  class MetaML < Meta
    def alt_key() "ml" end

    def initialize(ml)
      @alt_key = ml
    end

    def delete_all()
      my.query("delete from #{table} where #{alt_key}=#{q @alt_key}")
    end
  end

  class Member < MetaML
    def table() "member" end
    def key() "addr" end
    def items() [] end
    def log_str() "member" end
  end

  class Admin < MetaML
    def table() "admin" end
    def key() "admin" end
    def items() [] end
    def log_str() "admin" end
  end

  class Child < MetaML
    def table() "tree" end
    def key() "child" end
    def items() [] end
    def log_str() "child" end
    def alt_key() ["parent"] end
  end

  class Parent < MetaML
    def table() "tree" end
    def key() "parent" end
    def items() [] end
    def log_str() "parent" end
    def alt_key() ["child"] end
  end

  class Spool < MetaML
    def table() "spool" end
    def key() "seq" end
    def items() ["subject", "body"] end
    def log_str() "spool" end

    def spool(seq, subject, body)
      add(seq.to_s, {"subject"=>subject, "body"=>body})
    end

    def copy_messages(msgs, dest)
      sql = "select seq,body from spool where seq in(#{msgs.join(",")})"
      my.query(sql).each do |i,b|
        File.open("#{dest}/#{i}", "w"){|f| f.print b}
      end
    end

    def summary_list(msgs=nil)
      where = msgs ? "where seq in(#{msgs.join(",")})" : ""
      sql = "select seq,subject from spool #{where}"
      ret = ""
      my.query(sql).each do |i,s|
        ret << "#{i} #{s}\n"
      end
      ret
    end
  end

  class Work < MetaML
    def table() "work" end
    def key() "ident" end
    def items() ["data"] end

    def add(key, data)
      super(key, {"data"=>data})
    end

    def get(key)
      super(key)["data"]
    end

    def expire()
      my.query("delete from work where unix_timestamp(timestamp)+24*60*60 < unix_timestamp(NOW())")
    end
  end

end

# Test
#
if __FILE__ == $0 then
  require "tml.rb"
  def TML.log(a) puts a end

  $mysql_db = "tml"
  $mysql_user = "tml"
  $mysql_passwd = "tmlpwd"

  # Lock test
  l = TML::Mysql::Lock.new("hoge")
  begin
    TML::Mysql::Lock.new("hoge")
    raise "error!"
  rescue => e
    raise unless e.to_s == "dead lock detected: hoge"
  end
  l.unlock
  TML::Mysql::Lock.new("hoge").unlock

  # ML test
  ml = TML::Mysql::ML.new
  ml.add("hoge")
  raise "error!" unless ml.list == ["hoge"]
  ml.set("hoge", {"overview"=>"HOGEHOGE", "not_member"=>"true"})
  h = ml.get("hoge")
  raise "error!" if h["overview"] != "HOGEHOGE" or h["not_member"] != "true"

  # Counter test
  ct = TML::Mysql::Counter.new
  ct.add("hoge")
  raise "error!" unless ct.get("hoge") == 0
  ct.set("hoge", 99)
  raise "error!" unless ct.get("hoge") == 99
  ct.delete("hoge")
  raise "error!" if ct.have? "hoge"

  # Passwd test
  pw = TML::Mysql::Passwd.new
  pw.add("hoge", "hogepw")
  raise "error!" unless pw.list == ["hoge"]
  pw.auth("hoge", "hogepw")
  pw.set("hoge", "hogepw2")
  pw.auth("hoge", "hogepw2")
  pw.delete("hoge")
  raise "error!" if pw.have? "hoge"

  # Member test
  mm = TML::Mysql::Member.new("hoge")
  mm.add("hogehoge")
  raise "error!" unless mm.list == ["hogehoge"]
  mm.delete("hogehoge")
  raise "error!" if mm.have? "hogehoge"

  # Admin test
  ad = TML::Mysql::Admin.new("hoge")
  ad.add("xxxx")
  raise "error!" unless ad.list == ["xxxx"]
  ad.delete("xxxx")
  raise "error!" if ad.have? "xxxx"

  # Tree test
  ml.add("fuga")
  ch = TML::Mysql::Child.new("hoge")
  ch.add("fuga")
  raise "error!" unless ch.list == ["fuga"]
  raise "error!" unless TML::Mysql::Parent.new("fuga").list == ["hoge"]
  ch.delete("fuga")
  raise "error!" if ch.have? "fuga"
  ml.delete("fuga")
  ml.delete("hoge")
  raise "error!" if ml.have? "hoge"

end
